package com.nisum.hyd.service;

import com.mongodb.client.MongoClient;
import com.mongodb.client.MongoClients;
import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoDatabase;
import com.nisum.hyd.dto.UserCardDTO;
import org.apache.log4j.Logger;
import org.apache.spark.sql.*;
import org.apache.spark.sql.streaming.StreamingQuery;
import org.apache.spark.sql.streaming.StreamingQueryException;
import org.apache.spark.sql.types.StructType;

import org.bson.Document;
import org.springframework.stereotype.Service;

import java.util.concurrent.TimeoutException;

@Service
public class SparkKafkaConsumer {
    private static final Logger LOGGER = Logger.getLogger(SparkKafkaConsumer.class);
    public static final String MONGODB_DOCUMENT_COLLECTION = "mongodb://127.0.0.1/userdetailsdb.userdetails";
    public static final String MONGODB_URI = "mongodb://localhost:27017";
    private static final String KafkaBrokerEndpoint = "localhost:9092";
    private static String KafkaTopic = "UserWithCardDetails";

    public String getUserDetailsAndSaveToDB() {
        LOGGER.info("getUserDataAndSave() calling");
        SparkSession sparkSession = SparkSession.builder()
                .master("local[*]")
                .appName("MongoSparkConnector")
                .config("spark.mongodb.input.uri", MONGODB_DOCUMENT_COLLECTION)
                .config("spark.mongodb.output.uri", MONGODB_DOCUMENT_COLLECTION)
                .getOrCreate();

        Dataset<UserCardDTO> userDataset = getFilteredData(sparkSession);
        try {
            saveToDB(userDataset);
        } catch (Exception e) {
            LOGGER.error("Failed to write into database in getUserDataAndSave(): {}");
            e.printStackTrace();
        }
        //Dataset Saving into MongoDB without Streaming
        /*try {
            MongoSpark.save(userDataset);
        } catch (Exception e) {
            LOGGER.error("Failed to write into database in getUserDataAndSave(): {}");
            e.printStackTrace();
        }*/
        return "Data Saved in DB";
    }

    Dataset<UserCardDTO> getFilteredData(SparkSession sparkSession) {
        LOGGER.info("getFilteredData() calling");
        Dataset<Row> ds = sparkSession.readStream().format("kafka")
                .option("kafka.bootstrap.servers", KafkaBrokerEndpoint)
                .option("subscribe", KafkaTopic)
                .option("header", "true")
                .option("startingOffsets", "earliest")
                .load();
        StructType structType = Encoders.bean(UserCardDTO.class).schema();
        Dataset<Row> rowDataset = ds
                .selectExpr("CAST(value as String) as message")
                .select(functions.from_json(functions.col("message"), structType).as("userData"))
                .select("userData.*").distinct();
        rowDataset.show();
        Dataset<Row> filteredData = rowDataset.filter(functions.col("cardType").equalTo("Visa"));
        Dataset<UserCardDTO> productDataset = filteredData.as(Encoders.bean(UserCardDTO.class));
        productDataset.show();
        return productDataset;
    }

    void saveToDB(Dataset<UserCardDTO> userDataset) throws TimeoutException, StreamingQueryException {
        LOGGER.info("saveToDB() calling");
        StreamingQuery query = userDataset.writeStream().outputMode("append")
                .option("checkpointLocation", "checkPointLocation/streamingJob")
                .foreach(new ForeachWriter<UserCardDTO>() {
                    private MongoClient mongoClient;
                    private MongoDatabase database;
                    private MongoCollection<Document> collection;

                    @Override
                    public boolean open(long partitionId, long epochId) {
                        mongoClient = MongoClients.create(MONGODB_URI);
                        database = mongoClient.getDatabase("userdetailsdb");
                        collection = database.getCollection("userdetails");
                        return true;

                    }

                    @Override
                    public void process(UserCardDTO userCardDTO) {
                        Document document = new Document();
                        document.append("userId", userCardDTO.getUserId());
                        document.append("userName", userCardDTO.getUserName());
                        document.append("regMobileNo", userCardDTO.getRegMobileNo());
                        document.append("gender", userCardDTO.getGender());
                        document.append("cardType", userCardDTO.getCardType());
                        document.append("cardNumber", userCardDTO.getCardNumber());
                        document.append("cardStatus", userCardDTO.getCardStatus());
                        collection.insertOne(document);
                    }

                    @Override
                    public void close(Throwable errorOrNull) {
                        mongoClient.close();
                    }
                }).start();
        query.awaitTermination();
    }
}
